#!/usr/bin/env python

# Copyright (c) 2000-2001 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307
# USA.

"""Engine is the Narval interpreter"""

__revision__ = "$Id: Engine.py,v 1.56 2001/10/12 15:08:44 syt Exp $"

import time
import string
import os
import threading
import Queue

from narval.lib import *
from narval.bibal import *
from narval.services.EventScheduler import EventScheduler
from narval.services.SocketManager import SocketManager
from narval.services.RPC import RpcService
from narval.services.ComService import ComService
from narval.dom.ALDOMImplementation import ALImplementation
from narval.dom.ChangeListeners import *

# EngineThread #################################################################

class EngineThread(threading.Thread) :
    """EngineThread"""

    Engine = None

    def _Thread__stop(self) :
        self.Engine.end_of_thread()
        threading.Thread._Thread__stop(self)
        

class ThreadMgmtInterface :
    """Interface that defines the methods used to create and end EngineThreads
    """

    def create_thread(self,*args,**kargs) :
        raise Exception("interface method not implemented")

    def end_of_thread(self) :
        raise Exception("interface method not implemented")


# Narval Interfaces ############################################################

class MessageInterface :
    """Interface that defines Engine methods used for inter-Agent communication
    """

    def postMessage(self,messageXml) :
        raise Exception("interface method not implemented")
    
class DebugInterface :
    """Interface that defines Engine methods used for debug"""

    def debugSuspend(self) :
        raise Exception("interface method not implemented")
        
    def debugContinue(self) :
        raise Exception("interface method not implemented")

    def debugStepOneStep(self) :
        raise Exception("interface method not implemented")

class InteractionInterface :
    """Interface that defines Engine methods used by (Graph|CmdLine)UI"""

    def getMemory(self) :
        raise Exception("interface method not implemented")

    def getElement(self,elementId) :
        raise Exception("interface method not implemented")

    def addElement(self,elementXML) :
        raise Exception("interface method not implemented")

    def removeElement(self,elementId) :
        raise Exception("interface method not implemented")

    def replaceElement(self,elementId,newElementXml) :
        raise Exception("interface method not implemented")

    def startPlan(self,recipeName) :
        raise Exception("interface method not implemented")

    def saveRecipes(self) :
        raise Exception("interface method not implemented")

    def saveTransforms(self) :
        raise Exception("interface method not implemented")

    def shutdown(self) :
        raise Exception("interface method not implemented")

    ### ONLY USED BY HORN2_1 CVS BRANCH
    def getElementsList(self):
        raise Exception("interface method not implemented")

    def getElementList(self,elementId):
        raise Exception("interface method not implemented")

# Narval #######################################################################

class Narval(MemoryChangeListenerInterface, StepChangeListenerInterface,
             TransitionChangeListenerInterface, PlanChangeListenerInterface,
             ThreadMgmtInterface, MessageInterface, DebugInterface,
             InteractionInterface) :
    """
    Narval is your pal ! FIXME:OUTDATED

    Terminology:
    * memory: place where Narval stores all it knows of.
    * elements: pieces of information stored in memory
    * recipe: a recipe is an element that describes a sequence of steps,
      separated by transitions. A recipe does nothing. It is knowledge.
    * plan: a plan is an element. A plan is an instance of a recipe. A plan
      actually does something.
    * step: A step can point to an action or a recipe.
    * transition: a transition links steps in a recipe. it can be fired iff
      the associated condition is true.
    
    The clever part of Narval is supposed to be run_step() and eval_transition()
    because of the way they build the list of arguments to pass to the actions
    and deal with the output.

    Other advantages of Narval are supposed to be:
    * its reflexivity: it can reason about what it is doing, as what it is
      doing are running plans and plans are elements in memory. it can reason
      about what it knows how to do, as recipes are elements in memory. it can
      reason about what it can do as each action is defined as a transformation
      (set of inputs and outputs).
    * its possibilities for machine learning. because everything is recorded,
      and because of the above flexibility, lots of possibilities are opened for
      automatic learning.
    """

    # Init #####################################################################
    
    def __init__(self):

        # alive and for long
        self.alive = true        
        self.quit_when_done = false
        self.save_mem_on_quit = None
                
        # event queue
        self.event_queue = Queue.Queue(0)

        # thread management
        self.thread_queue = Queue.Queue(0)
        self.max_threads = 10
        EngineThread.Engine = self
        
        # debug
        self.step_by_step = false
        self.step_by_step_semaphore = threading.Semaphore(0)
                
        # create memory document
        self.memoryDocument = ALImplementation.createMemoryDocument(self)
        self.memory = self.memoryDocument.documentElement
        
        # create scheduler
        self.eventScheduler = EventScheduler(self.post_event)
        
        # services
        self.socketManager = None
        self.rpcServer = None
        self.comService = None

        # debug mode
        self.debug = 0
    # Services #################################################################

    def _start_rpc(self,port) :
        if not self.rpcServer :
            self._start_comService()
            self.rpcServer = RpcService(port,self,self.comService)
            self.rpcServer.start()

    def _stop_rpc(self) :
        if self.rpcServer :
            self.rpcServer.stop()

    def _start_socketManager(self) :
        if not self.socketManager :
            self.socketManager = SocketManager()
            self.socketManager.start()
    
    def _stop_socketManager(self) :
        if self.socketManager :
            self.socketManager.stop()

    def _start_comService(self) :
        if not self.comService :
            self.comService = ComService()
            self.comService.start()

    def _stop_comService(self) :
        if self.comService :
            self.comService.stop()
            
    # (Un)Load stuff in memory #################################################

    def _load_recipes(self) :
        loadRecipes(self.memoryDocument, self.memoryDocument.add_element)
        ALImplementation.checkpointMemory()

    def _save_recipes(self) :
        errors = saveRecipes(self.memory)
        self._process_file_saving_errors(errors)
            
    def _load_modules(self) :
        loadModules(self.memoryDocument, self.memoryDocument.add_element)
        ALImplementation.checkpointMemory()
    
    def _load_transforms(self) :
        loadTransforms(self.memoryDocument, self.memoryDocument.add_element)
        ALImplementation.checkpointMemory()

    def _save_transforms(self) :
        errors = saveTransforms(self.memory)
        self._process_file_saving_errors(errors)
    
    def _load_memory(self,path=None) :
        loadMemory(self.memoryDocument, self.memoryDocument.add_element, path)
        ALImplementation.checkpointMemory()

    def _save_memory(self,path) :
        write_domtree_to_file(self.memoryDocument.documentElement,path,encoding='UTF-8')
        
    def _process_file_saving_errors(self,errors):
        if errors:
            d = self.memoryDocument
            err = d.createElementNS(AL_NS,'error')
            m = d.createElementNS(AL_NS,'message')
            m.appendChild(d.createTextNode('The following files could not be saved. Please check that you have write access.'))
            err.appendChild(m)
            for e in errors:
                f = d.createElementNS(AL_NS,'filename')
                f.appendChild(d.createTextNode(e))
                err.appendChild(f)
            d.add_element(err)

    # Event Queue & Schedule ###################################################

    def post_event(self, event) :
        self._fire_event('engine','postQueueEvent', str(id(event)), str(event))
        self.event_queue.put(event)

    def _pop_event(self) :
        ev = self.event_queue.get()
        self._fire_event('engine','popQueueEvent', str(id(ev)))
        return ev
        
    def schedule_event(self,event,when=0,period=0,absoluteDate=0):
        ev = (event,when,period,absoluteDate)
        self._fire_event('engine','postSchedulerEvent', str(id(ev)), str(ev))
        self.eventScheduler.schedule_event(ev)

    # Run, quit and Process ####################################################

    def run(self) :
        """Mainloop. Execute plans endlessly unless a debugger says 'wait'.
        """
        print '\n"Hello, I\'m Narval your pal !"'
        self.eventScheduler.start()
        while self.alive :
            # done ?
            if self.quit_when_done and self.event_queue.empty() \
               and not self.memoryDocument.are_active_plans() :
                self._quit()
                continue

            # clear lists
            self.ready_steps = []
            self.ready_plans = []

            # debug
            if self.step_by_step :
                self.step_by_step_semaphore.acquire()

            # wait for event
            event = self._pop_event()
            self._process_event(event)

            # fire transitions
            #map(lambda plan: plan.run(), self.ready_plans)
            [plan.run() for plan in self.ready_plans]
            # run actions
            #map(lambda step: step.run(), self.ready_steps)
            [step.run() for step in self.ready_steps]

    def _quit(self) :
        print "I'm dying: event quit in the queue ! <:~("
        self.alive = false

        # stop services
        self.eventScheduler.stop()
        self._stop_rpc()
        self._stop_socketManager()
        self._stop_comService()

        # save memory to file ?
        if self.save_mem_on_quit :
            self._save_memory(self.save_mem_on_quit)
            
        # get rid of documents
        if ALImplementation.tx:
            ALImplementation.tx.commit()
        else:
            release_domtree(self.memoryDocument)
            del self.memoryDocument        


    def _process_event(self,event) :
        if event[0] == 'endOfThread' :
            self._run_thread()

        # start plan ('startPlan', planName,(parent_plan,parent_step),elements)
        elif event[0] == 'startPlan' :
            self._start_plan(event[1],event[2],event[3])
            
        # planEnd ('planEnd', plan)
        elif event[0] == 'planEnd' :
            event[1].end()

        # stepEnd ('stepEnd', step)
        elif event[0] == 'stepEnd' :
            event[1].end()
        
        # quitting
        elif event[0] == 'quit' :
            self._quit()

        # postMessage ('postMessage', message_as_xml)
        elif event[0] == 'postMessage' :
            self.memoryDocument.add_message_as_string(event[1])

        # addElement ('addElement', element_as_xml)
        elif event[0] == 'addElement' :
            self.memoryDocument.add_element_as_string(event[1])
        
        # removeElement ('removeElement', eid)
        elif event[0] == 'removeElement' :
            self.memoryDocument.remove_element_by_id(event[1])
        
        # replaceElement ('replaceElement', eid, element_as_xml)
        elif event[0] == 'replaceElement' :
            self.memoryDocument.replace_element_as_string(event[1],event[2])
        
        # time condition ('time-condition', transition)
        elif event[0] == 'time-condition' :
            event[1].time_condition_match()

        # plan decay ('forgetPlan', plan)
        elif event[0] == 'forgetPlan' :
            self.memoryDocument.remove_element(event[1])

        # start/stop rpc ('startRPC', port) ('stopRPC')
        elif event[0] == 'startRPC' :
            self._start_rpc(event[1])
        elif event[0] == 'stopRPC' :
            self._stop_rpc()

        # start/stop socketManager ('startSocketManager') ('stopSocketManager')
        elif event[0] == 'startSocketManager' :
            self._start_socketManager()
        elif event[0] == 'stopSocketManager' :
            self._stop_socketManager()

        # load memory/recipes/modules/transforms
        elif event[0] == 'loadMemory' :
            self._load_memory(event[1])
        elif event[0] == 'loadRecipes' :
            self._load_recipes()
        elif event[0] == 'loadModules' :
            self._load_modules()
        elif event[0] == 'loadTransforms' :
            self._load_transforms()        
        
        # unknown
        else :
            print "WARNING: UNKNOWN EVENT",event

    # Start plan ###############################################################

    def _start_plan(self,planName,parents,elements) :
        """Change plan and first step states, then run first step
        """
        recipe = self.memoryDocument.get_recipe(planName)
        if recipe :
            code, msg = recipe.check_syntax()
        else :
            msg = 'Unknown recipe %s' % (planName)
                
        if msg :
            error = self.memoryDocument.mk_error('Bad recipe', msg)
            self.memoryDocument.add_element(error)
            print error
            print msg
            return

        # make plan and add it to memory
        plan = recipe.make_plan()
        self.memoryDocument.add_element( plan )
        plan.add_elements(elements)
        plan.set_parents(parents)

        # fire event
        self._fire_event('plan','instanciateRecipe',
                         plan.getAttributeNS(AL_NS,'eid'),
                         recipe.getAttributeNS(AL_NS,'name'))

        # run first step
        plan.start()

    # Thread management ########################################################

    def create_thread(self,*args,**kargs) :
        self.thread_queue.put( (args,kargs) )
        self._run_thread()

    def _run_thread(self) :
        if self.debug:
            print threading.activeCount(),'threads running...'
        if threading.activeCount() < self.max_threads :
            try:
                args, kargs = self.thread_queue.get_nowait()
                t = apply(EngineThread,args,kargs)
                t.start()
                if self.debug:
                    print 'started new thread'
            except Queue.Empty :
                pass

    def end_of_thread(self) :
        if self.debug :
            print 'thread done'
        self.post_event( ('endOfThread',) )
        
    # Internal events ##########################################################

    # take care to do stuff *before* posting events to the event queue or the
    # thready-boogieman will catch you by processing the event before you get
    # time to blink !
    
    def memory_change(self,action,element,old_element=None) :
        eid = element.getAttributeNS(AL_NS,'eid')
        ASSERT(eid,'No eid in Memory Element ')
        # fire event
        if action == 'add' :
            self._fire_event('memory','addElement',eid)
        elif action == 'remove' :
            self._fire_event('memory','removeElement',eid)
        elif action == 'replace':
            self._fire_event('memory','replaceElement',eid)
        
        # test
        # be careful : the order is important, esp. in case we replace
        # a socket by another socket
        if action != 'remove' and element.tagName == 'start-plan' :
            self.post_event( ('startPlan',element.getAttributeNS(AL_NS,'recipe'),
                             (None,None),[]) )

        if self.socketManager:
            if (element.tagName in ['server-socket','socket']) and \
               (action in ['remove','replace']) :
                socket_port = int(element.getAttributeNS(AL_NS,'port'))
                self.socketManager.kill_socket(socket_port)
                
            if action != 'remove' and element.tagName == 'server-socket' :
                self.socketManager.new_server_socket( 
                    int(element.getAttributeNS(AL_NS,'port')),
                    self.sm_cbk_srv_createplan,
                    element.getAttributeNS(AL_NS,'recipe'))

            if element.tagName == 'client-socket':
                h = element.getAttributeNS(AL_NS,'host')
                p = int(element.getAttributeNS(AL_NS,'port'))
                if action in ('remove','replace'):
                    sock_elt = select_one(self.memoryDocument,
                                          'socket[@host=%s and @port=%d]'%(host,port))
                    self.memoryDocument.remove_element(sock_elt)
                    self.socketManager.kill_socket((h,p))
                if action !='remove':
                    self.socketManager.new_client_socket( 
                        h,p,
                        self.sm_cbk_cli_socket_element_creator,
                        self.sm_cbk_cli_createplan,
                        element.getAttributeNS(AL_NS,'recipe'))
                        
        elif element.tagName[-6:]=='socket' :
            print 'No socketManager running !'
                

    def step_change(self,step) :
        state = step.getAttributeNS(AL_NS,'state')
        self._fire_event('plan','stepStatusChanged',
                         step.parentNode.getAttributeNS(AL_NS,'eid'),
                         step.getAttributeNS(AL_NS,'id'),
                         state)
        if state == 'ready' :
            self.ready_steps.append(step)
        elif state == 'end' :
            self.post_event( ('stepEnd',step) )


    def plan_change(self,plan,action='state',element=None) :
        if action == 'state' :
            state = plan.getAttributeNS(AL_NS,'state')
            self._fire_event('plan','planStatusChanged',
                             plan.getAttributeNS(AL_NS,'eid'),
                             state)
            if state == 'ready' :
                self.ready_plans.append(plan)
            elif state in ['end','failed-end'] :
                self.post_event( ('planEnd',plan) )
            elif state in ['done','failed']:
                # schedule decay
                self.schedule_event(('forgetPlan',plan),
                                   int(plan.getAttributeNS(AL_NS,'decay')))
        elif action == 'add' :
            self._fire_event('plan','planAddElement',
                             plan.getAttributeNS(AL_NS,'eid'),
                             element.getAttributeNS(AL_NS,'eid'))
        elif action == 'remove' :
            self._fire_event('plan','planRemoveElement',
                             plan.getAttributeNS(AL_NS,'eid'),
                             element.getAttributeNS(AL_NS,'eid'))


    def transition_change(self,transition) :
        self._fire_event('plan','transitionStatusChanged',
                         transition.parentNode.getAttributeNS(AL_NS,'eid'),
                         transition.getAttributeNS(AL_NS,'id'),
                         transition.getAttributeNS(AL_NS,'state'))

    # Socket Manager callbacks #################################################
        
    def sm_cbk_srv_createplan(self,port,recipe_name) :
         """This function is called by the SocketManager when it receives a
         connection on a server-socket. This function's role is to start the
         appropriate plan with a socket element in its memory
         """
         sock_elt = self.memoryDocument.createElementNS('','socket')
         sock_elt.setAttributeNS(AL_NS,'port',str(port))
         self.post_event(('startPlan',recipe_name,(None,None),[sock_elt]))

    def sm_cbk_cli_createplan(self,port,recipe_name) :
         """This function is called by the SocketManager when it receives a
         connection on a client-socket. This function's role is to start the
         appropriate plan with a socket element in its memory
         """
         print '/*/socket[@local_port="%d"]'%(port)
         sock_elt = select_one(self.memoryDocument,'/*/socket[@local_port="%d"]'%(port))
         self.post_event(('startPlan',recipe_name,(None,None),[sock_elt]))

    def sm_cbk_cli_socket_element_creator(self,host,port,local_port):
         sock_elt = self.memoryDocument.createElementNS(AL_NS,'socket')
         sock_elt.setAttributeNS(AL_NS,'port',str(port))
         sock_elt.setAttributeNS(AL_NS,'host',host)
         sock_elt.setAttributeNS(AL_NS,'local_port',str(local_port))
         self.memoryDocument.add_element(sock_elt)
    
    # fire Event ###############################################################

    def _fire_event(self,target,method,*args):
        if self.comService :
            self.comService.add_event(target, method, args)

    # Memory manipulation ######################################################

    def addElement(self,elementXml):
        self.post_event( ('addElement',elementXml) )

    def removeElement(self,elementId):
        self.post_event( ('removeElement',elementId) )

    def startPlan(self,recipeName):
        self.post_event( ('startPlan', recipeName, (None,None),[]) )

    def replaceElement(self,elementId,newElementXml):
        self.post_event( ('replaceElement',elementId,newElementXml) )

    def getMemory(self):
        return element_to_string(self.memory,false,'UTF-8')

    def getElement(self,elementId):
        try:
            element = select_one(self.memory,'*[@eid="%s"]'%(elementId))
            return element_to_string(element,false,'UTF-8')
        except SelectException :
            return None

    ### ONLY USED BY HORN2_1 CVS BRANCH
    def getElementsList(self):
        l = []
        for n in self.memory.childNodes:
            l.append(self._get_list(n))
        return l
    def getElementList(self,elementId):
        try:
            element = select_one(self.memory,'*[@eid="%s"]'%(elementId))
            return self._get_list(element)
        except SelectException :
            return None
    def _get_list(self, element):
        return (element.getAttributeNS('', 'eid'), element.nodeName,
                element.getAttributeNS('', 'group'),
                element.getAttributeNS('', 'name'))
    
    # Messaging ################################################################

    def postMessage(self,messageXml) :
        self.post_event( ('postMessage',messageXml) )
        
    # Debug ####################################################################

    def debugSuspend(self):
        self.step_by_step = true

    def debugContinue(self):
        self.step_by_step = false
        self.step_by_step_semaphore.release()

    def debugStepOneStep(self):
        if self.step_by_step :
            self.step_by_step_semaphore.release()
        return self.step_by_step

    # RemoteControl ############################################################

    def saveRecipes(self) :
        self._save_recipes()

    def saveTransforms(self) :
        self._save_transforms()

    def shutdown(self) :
        self.post_event( ('quit',) )
        
# Start / Stop #################################################################

def start(narval,manageSignals = 0) :
    """Start Narval"""

    if manageSignals :
        print "manage signals"
        import signal
        signal.signal(signal.SIGINT, lambda x,y,z = narval: stop(z))
        signal.signal(signal.SIGTERM, lambda x,y,z = narval: stop(z))

    narval.run()
    print "Narval is dead..."


def stop(narval) :
    print '################### INTERRUPT'
    narval.post_event( ('quit',) )


# Main #########################################################################

USAGE = """Usage: narval [options]
   --debug           : print plan and memory messages
   --home DIR        : overwrite NARVAL_HOME environment variable with DIR
   --quit-when-done  : quit when no more running plan (only done or failed)
   --quit-after N    : quit after N seconds of execution
   --start-plan PLAN : execute PLAN right after starting
   --listen-on PORT  : listens for UI client connections on specified port
   --socket-manager  : start socket manager
   --max-threads N   : sets maximum number of concurrent threads to N
   --load-memory-file: uses specified FILE instead of default memory file
   --save-mem-on-quit: save memory to FILE when quitting
   --help            : print this help message
   """

def manage_home(path, chdir=1) :
    from narval import config
    if not config.set_home(path):
        msg = ('The home that you selected does not exist, or does not '+\
               'contain the required subdirectories. I can:\n'+\
              '\t[C]reate an empty directory structure for you\n'+\
              '\t[P]opulate a new directory structure using data from %s\n'+\
              '\t[S]top execution.') % config.get_share()
        print msg
        answer = ''
        while answer not in ['c','p','s']:
            answer = string.lower(raw_input('--> '))
        if answer == 'c':
            config.create_home_subdirs()
        elif answer == 'p':
            check_for_home()
        else:
            sys.exit(0)
    print 'Using %s as home' % path
    if chdir:
        os.chdir(path)
    
def Run(sysArgs):
    # process options
    import getopt
    short_list = ''
    long_list = ['debug', 'quit-after=', 'quit-when-done', 'start-plan=',
                 'home=', 'help', 'use-database','listen-on=',
                 'socket-manager', 'max-threads=',
                 'load-memory-file=', 'save-mem-on-quit=']
    optlist, args = [], []
    try:
        optlist, args = getopt.getopt(sysArgs[1:], short_list, long_list)
    except getopt.error,e :
        print 'Warning:',e
        print USAGE
        sys.exit(1)

    # specific home or memory setup ?
    mempath, home = None, None
    for opt, val in optlist :
        if opt =='--home' :
            home = os.path.abspath(val)
        elif opt == '--use-database' :
            print 'setting use_database to 1'
            ALImplementation.use_database = 1
        elif opt == '--load-memory-file' :
            print 'using',val,'as memory file'
            mempath = os.path.abspath(val)
    # FIXME manage_home even with default home ?
    if home:
        manage_home(home)
    # make sure that NARVAL_HOME exist and is populated
    check_for_home()
    
    # create Narval instance
    n = Narval()
    n.post_event( ('loadRecipes',) )
    n.post_event( ('loadTransforms',) )
    n.post_event( ('loadModules',) )
    n.post_event( ('loadMemory',mempath) )

    # process options
    for opt, val in optlist :
        if opt == '--debug' :
            n.debug = 1
            n._start_comService()
            from narval.PlanListener import *
            from narval.MemoryListener import *
            n.comService.addPlanListener(PlanListenerPrintImplementation())
            
            n.comService.addMemoryListener(MemoryListenerPrintImplementation())

        elif opt == '--quit-after' :
            n.schedule_event(('quit',),when=int(val))

        elif opt == '--quit-when-done' :
            n.quit_when_done = true

        elif opt == '--save-mem-on-quit' :
            n.save_mem_on_quit = val

        elif opt == '--start-plan' :
            n.post_event( ('startPlan',val,(None,None),[]) )

        elif opt == '--listen-on':
            n.post_event( ('startRPC',int(val),(None,None),[]) )
            from socket import gethostname
            n.addElement("<listen-on host='%s' port='%d'/>"%(gethostname(),int(val)))
            
        elif opt == '--socket-manager':
            n._start_socketManager()

        elif opt == '--max-threads' :
            n.max_threads = int(val)
            
        elif opt in ['--home','--use-database','--load-memory-file'] :
            pass

        else :
            print USAGE
            sys.exit(1)
            
    # start
    start(n,0)

if __name__ == "__main__" :
    import sys
    Run(sys.argv)
