# Copyright (c) 2000-2001 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""MemoryDocument module"""

__revision__ = "$Id: MemoryDocument.py,v 1.37 2001/09/03 10:09:34 alf Exp $"

from ALDomAbstraction import ALElement, ALDocument
from ActionElement import ActionElement
from PlanElement import PlanElement
from RecipeElement import RecipeElement
from StepElement import StepElement
from TransitionElement import TransitionElement
from TransformElement import TransformElement
from ChangeListeners import *
from Tags import *
from Match import match_filter

from narval.bibal import *
from narval.lib import *

import string

class MemoryDocument(ALDocument, TransitionChangeListenerInterface,
                     StepChangeListenerInterface, PlanChangeListenerInterface) :
    """MemoryDocument XXX"""

    # Document methods overriden ###############################################
    
    def __init__(self,al=None) :
        try:
            ALDocument.__init__(self,None)
        except:
            ALDocument.__init__(self)
        self.Al = al
        self.eid_count = 1
        self.eid_ref_count = {}
        self.actions = {}
        self.transforms = {}
        self.recipes = {}
        self.eid_mapping = {}
        
    def createElement(self,tagName) :
        return self.createElementNS('',tagName)

    def createElementNS(self, namespaceURI, qualifiedName):
        fields = string.split(qualifiedName, ':')
        if len(fields) == 2:
            prefix = fields[0]
            localName = fields[1]
        elif len(fields) == 1:
            prefix = ''
            localName = fields[0]
        #FIXME: I need to look up the full list of "not allowed" chars
        for c in localName:
            if c in ['<','>','&']:
                raise DOMException(INVALID_CHARACTER_ERR)

        if qualifiedName == 'recipe' :
            e = RecipeElement(self, qualifiedName, namespaceURI, prefix, localName)
        elif qualifiedName == 'plan' :
            e = PlanElement(self, qualifiedName, namespaceURI, prefix, localName)
        elif qualifiedName == 'transition' :
            e = TransitionElement(self, qualifiedName, namespaceURI, prefix, localName)
        elif qualifiedName == 'step' :
            e = StepElement(self, qualifiedName, namespaceURI, prefix, localName)
        elif qualifiedName == 'action' :
            e = ActionElement(self, qualifiedName, namespaceURI, prefix, localName)
        elif qualifiedName == 'transform' :
            e = TransformElement(self, qualifiedName, namespaceURI, prefix, localName)
        else :
            e = ALElement(self, qualifiedName, namespaceURI, prefix, localName)
        return e

    # Memory methods ###########################################################
    
    def add_elements(self,elements) :
        map(self.add_element,elements)


    def add_element(self,element) :
        """if the element has not already an eid (== not yet in memory)
        assign unique id to the element and append it to memory.
        """
        eid = element.getAttributeNS(AL_NS,'eid')
        if eid :
            self.eid_ref_count[eid] = self.eid_ref_count[eid] + 1
        else :
            # assign new eid to element
            eid = str(self.eid_count)
            element.setAttributeNS(AL_NS,'eid',eid)
            element.setAttributeNS(AL_NS,'timestamp',mk_time_stamp())
            self.eid_count = self.eid_count + 1
            self.eid_ref_count[eid] = 1
            # append to memory document
            self.documentElement.appendChild(element)
            # do some caching processing
            self.eid_mapping[eid]=element
            name = element.nodeName
            if name == 'recipe':
                self.recipes[element.get_identity()]=eid
            if name == 'action':
                self.actions[element.get_identity()]=eid
            if name == 'transform':
                self.transforms[element.get_identity()]=eid
            
            # if debug mode, and element is an error, PrettyPrint it to stdout
            if self.Al.debug and element.nodeName == 'error' :
                PrettyPrint(element)

            # element added to memory: fire event.
            self.Al.memory_change('add',element)

            # this new element may trigger a transition. propagate.
            for node in self.documentElement.childNodes[:] :
                if node.tagName == 'plan' :
                    node.element_change(element)

        return eid, element


    def add_message_as_string(self,msg_str) :
        element = read_domtree_from_string(msg_str,self).firstChild
        element.setAttributeNS(AL_NS,'type','incoming')
        if element.tagName == 'message' :
            self.add_element(element)
        else :
            print "Warning: received a false message. Dropped it."

        
    def add_element_as_string(self,element_str) :
        element = read_domtree_from_string(element_str,self).firstChild
        self.add_element(element)

        
    def remove_element(self,element) :
        if element.getAttributeNS(AL_NS,'persist') == 'yes' and \
           element.getAttributeNS(AL_NS,'outdated') != 'yes' :
            return
        self.documentElement.removeChild(element)
        clear_tags(element)
        self.Al.memory_change('remove',element)
        name = element.tagName
        del self.eid_mapping[element.getAttributeNS(AL_NS,'eid')]
        if name == 'recipe':
            del self.recipes[element.get_identity()]
        elif name == 'action':
            del self.actions[element.get_identity()]
        elif name == 'transform':
            del self.transforms[element.get_identity()]
        elif name == 'plan' :
            element.remove_elements(element.elements)
        if name in ('plan','recipe'):
            for e in element.childNodes:
                if e.tagName in ['step','transition'] :
                    e.unlink()
            element.unlink()
            
        release_domtree(element)


    def remove_element_by_id(self,eid) :
        try:
            self.remove_element(self.eid_mapping[eid])
        except KeyError :
            print 'Warning: No element with eid',eid,'in memory'


    def replace_element(self,old_element,new_element):
        new_element.setAttributeNS(AL_NS,'eid',old_element.getAttributeNS(AL_NS,'eid'))
        self.documentElement.replaceChild(new_element,old_element)
        self.eid_mapping[old_element.getAttributeNS(AL_NS,'eid')]=new_element
        oldName = old_element.tagName
        newName = new_element.tagName
        if oldName != newName :
            if oldName == 'recipe':
                del self.recipes[old_element.get_identity()]
            elif oldName == 'action':
                del self.actions[old_element.get_identity()]
            elif oldName == 'transform':
                del self.transforms[old_element.get_identity()]
            if newName == 'recipe':
                self.recipes[new_element.get_identity()]=new_element
            elif newName == 'action':
                self.actions[new_element.get_identity()]=new_element
            elif newName == 'transform':
                self.transforms[new_element.get_identity()]=new_element
            
        self.Al.memory_change('replace',new_element,old_element)
        release_domtree(old_element)

    def replace_element_as_string(self,eid,element_str) :
        try:
            element = self.eid_mapping[eid]
            new_element = read_domtree_from_string(element_str,self).firstChild
            self.replace_element(element,new_element)
        except KeyError :
            print 'Warning: No element with eid',eid,'in memory'


    def delete_ref_to_element(self,element) :
        eid = element.getAttributeNS(AL_NS,'eid')
        self.eid_ref_count[eid] = self.eid_ref_count[eid] - 1
        if self.eid_ref_count[eid] < 1 :
            self.remove_element(element)
            
        
    def add_to_context(self, add_list, plan) :
        to_add = []
        memory = self.documentElement.childNodes[:]
        for add in add_list :
            rules = [x.data for x in select_many(add,'match/text()')]
            frm = add.getAttributeNS(AL_NS,'from')
            debug_msg('al_fetch','add_to_context %s %s %s'%(`add`,frm,`rules`))
            if frm == 'memory' :
                for e in memory :
                    if match_filter(e,rules) :
                        to_add.append(e)
            elif frm == 'plan' :
                for e in plan.elements :
                    if match_filter(e,rules) :
                        to_add.append(e)
            else :
                raise Exception("Unknown 'from' in add-to-context")
        return to_add


    def get_elements(self) :
        return self.documentElement.childNodes[:]


    def are_active_plans(self) :
        for node in self.documentElement.childNodes :
            if node.tagName == 'plan' and \
               node.getAttributeNS(AL_NS,'state') not in ['done','failed'] :
                return true
        return false
    

    def mk_error(self,type,msg) :
        error = self.createElementNS(AL_NS,'error')
        error.setAttributeNS(AL_NS,'type',type)
        error.appendChild(self.createTextNode(msg))
        return error


    def get_recipe(self, recipeName) :
        try:
            group, name = string.split(recipeName,'.')
            return self.eid_mapping.get(self.recipes.get((group,name)))
        except ValueError :
            return None
            

    def get_transform(self, name) :
        try:
            group, name = string.split(name,'.')
            return self.eid_mapping.get(self.transforms.get((group,name)))
        except ValueError :
            return None

    def get_action(self, name) :
        try:
            group, name = string.split(name,'.')
            return self.eid_mapping.get(self.actions.get((group,name)))
        except ValueError :
            return None

    # Proxy for Engine #########################################################

    def step_change(self,*args,**kargs) :
        return apply(self.Al.step_change,args,kargs)

    def plan_change(self,*args,**kargs) :
        return apply(self.Al.plan_change,args,kargs)

    def transition_change(self,*args,**kargs) :
        return apply(self.Al.transition_change,args,kargs)

    def transition_wait_time(self,transition,date):
        self.Al.schedule_event(('time-condition',transition),when=date,absoluteDate=1)

    def start_plan(self,recipe_name,parent_plan,parent_step,elements) :
        self.Al.post_event( ('startPlan', recipe_name,
                             (parent_plan, parent_step), []) )

    def create_thread(self,*args,**kargs) :
        apply(self.Al.create_thread,args,kargs)
        


