# Copyright (c) 2000-2001 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""PlanElement module"""

__revision__ = "$Id: PlanElement.py,v 1.43 2001/09/27 16:37:01 syt Exp $"


from narval.lib import *
from narval.bibal import *

from ALDomAbstraction import ALElement
from ALDOMImplementation import ALImplementation
from StepElement import StepElement
from TransitionElement import TransitionElement
from ChangeListeners import *

class PlanElement(ALElement, TransitionChangeListenerInterface,
                  ElementChangeListenerInterface, StepChangeListenerInterface) :
    """PlanElement represents a running Plan in Narval's Memory."""

    # State diagram Definition #################################################

    SDD = {'running': ['ready', 'failed-end', 'end'],
           'ready': ['running','failed-end'],
           'failed-end': ['failed'],
           'failed': [],
           'end': ['done'],
           'done': []
           }

    # Element methods overload #################################################

    def __repr__(self):
        #return "<PlanElement Node at %s: with %d attributes and %d children>" \
        #       % ( id(self), len(self.attributes), len(self.childNodes) )
        return "<PlanElement at %s eid='%s' name='%s'>" % \
               ( id(self), self.getAttributeNS(AL_NS,'eid'),
                 self.getAttributeNS(AL_NS,'name') )

    # Init methods #############################################################
    
    def __init__(*args): #self, ownerDocument, nodeName, namespaceURI, prefix, localName) :
        if len(args)==5:
            (self, ownerDocument, namespaceURI, prefix, localName) = args
            apply(ALElement.__init__, args)
        else:
            (self, ownerDocument, nodeName, namespaceURI, prefix, localName) = args
            ALElement.__init__(self,ownerDocument,nodeName,namespaceURI,prefix,localName)
        self.elements_node = self.ownerDocument.createElementNS(AL_NS,'elements')
        self.appendChild(self.elements_node)
        
        # document (used by the StepElements)
        self.document = ALImplementation.createPlanDocument()
##        print 'createdDocument',self,self.document


    def unlink(self) :
        """break cycles to facilitate garbage collection"""

        # start/end steps
        self.start_step = None
        self.end_step = None
        
        # transitions
        self.transitions = {'impossible':[],
                            'wait-step':[],
                            'wait-condition':[],
                            'wait-time':[],
                            'fireable':[],
                            'history':[]}

        # elements
        self.elements = []
        self.elements_node = None
        self.document = None

        # parent plan/step
        self.parent_plan = None
        self.parent_step = None
        self.init = false

        
    def link(self) :
        if self.init : return
        start = self.getAttributeNS(AL_NS,'start_step')
        end = self.getAttributeNS(AL_NS,'end_step')
        for node in self.childNodes :
            if node.__class__ == StepElement :
                node.setAttributeNS(AL_NS,'state','todo')
                node.link()
                if node.getAttributeNS(AL_NS,'id') == start :
                    self.start_step = node
                if node.getAttributeNS(AL_NS,'id') == end :
                    self.end_step = node
            elif node.__class__ == TransitionElement :
                node.setAttributeNS(AL_NS,'state','wait-step')
                node.link()
        self.init = true

    # Plan element methods #####################################################

    def set_parents(self,parents) :
        if parents[0] :
            self.parent_plan = parents[0]
            self.setAttributeNS(AL_NS,'parent_plan',
                                self.parent_plan.getAttributeNS(AL_NS,'eid'))
        if parents[1] :
            self.parent_step = parents[1]
            self.setAttributeNS(AL_NS,'parent_step',
                                self.parent_step.getAttributeNS(AL_NS,'id'))

    
    def start(self) :
        self.start_step.prepare()


    def run(self) :
        # look for transition of highest priority
        transition, priority = None, -1
        for t in self.transitions['fireable'] :
            try:
                p = int(t.getAttributeNS(AL_NS,'priority'))
            except ValueError :
                p = 0
            if p > priority :
                transition = t
                priority = p
        if not transition : assert 0, "Couldn't select transition to fire"

        # fire transition
        transition.fire()
        self.transitions['fireable'] = []


    def end(self) :
        # delete sockets and temp document
        sockets = [x.tagName == 'socket' for x in self.elements]
        self.remove_elements(sockets)
        release_domtree(self.document)
        self.document = None

        # change state
        state = self.getAttributeNS(AL_NS,'state')
        if state == 'end' :
            self._set_state('done')
        elif state == 'failed-end' :
            self._set_state('failed')

        # if plan has mark 'restart': restart it !
        if self.getAttributeNS(AL_NS,'restart') == 'yes' :
            planName = self.getAttributeNS(AL_NS,'group')+'.'+\
                       self.getAttributeNS(AL_NS,'name')
            self.ownerDocument.start_plan( planName, None, None, [] )


    def add_elements(self,elements) :
        """If element is not already in the context of the plan, add it.
        """
        difference = [elt for elt in elements if elt not in self.elements]
        for e in difference :
            eid, e = self.ownerDocument.add_element(e)
            self.elements.append(e)
            # create child of plan/elements
            elt = self.ownerDocument.createElementNS(AL_NS,'element')
            elt.setAttributeNS(AL_NS,'eid',e.getAttributeNS(AL_NS,'eid'))
            self.elements_node.appendChild(elt)
            # propagate happy news
            self.ownerDocument.plan_change(self,'add',e)
                
        if self.parent_plan :
            self.parent_plan.add_elements(elements)


    def remove_elements(self,elements) :
        intersection = [elt for elt in elements if elt in self.elements]
        for e in intersection :
            eid  = e.getAttributeNS(AL_NS,'eid')
            found = FALSE
            for elt in self.elements_node.childNodes:
                if elt.nodeName == 'element' and elt.getAttributeNS(AL_NS,'eid') == eid:
                    found = TRUE
                    self.elements_node.removeChild(elt)
                    self.elements.remove(e)
                    # propagate sad news
                    self.ownerDocument.plan_change(self,'remove',e)
                    # erase
                    self.ownerDocument.delete_ref_to_element(e)
                    release_domtree(elt)
                    break
            if not found :
                print "Element eid=",e.getAttributeNS(AL_NS,'eid'),'not in plan'

    # ChangeListeners interfaces ###############################################
    
    def transition_change(self,transition) :
        if transition.getAttributeNS(AL_NS,'state') == 'fireable' :
            self.transitions['fireable'].append(transition)
            self._set_state('ready')

        
    def step_change(self,step) :
        if step == self.end_step :
            s = step.getAttributeNS(AL_NS,'state')
            if s == 'done':
                self._set_state('end')
            elif s in ['impossible','failed','error'] :
                self._set_state('failed-end')
        if step.getAttributeNS(AL_NS,'state') == 'running' :
            self._set_state('running')
            
            
    def element_change(self,element) :
        for node in self.childNodes[:] :
            if node.tagName == 'transition' :
                node.element_change(element)
        
    # Private ##################################################################
    
    def _set_state(self,state) :
        # -- precondition
        oldstate = self.getAttributeNS(AL_NS,'state')
        if state == oldstate : return
        assert (state in self.SDD[oldstate]), \
               'Bad state diagram transition in plan %s from %s to %s' \
               % (self.getAttributeNS(AL_NS,'name'),oldstate,state)

        # change state
        self.setAttributeNS(AL_NS,'state',state)

        # propagate change
        self.ownerDocument.plan_change(self)
        if self.parent_step :
            self.parent_step.plan_change(self)

        # -- postcondition
        assert (self.getAttributeNS(AL_NS,'state') == state), \
                'Bad state %s after transition in plan %s from %s to %s' \
                % (self.getAttributeNS(AL_NS,'state'),
                   self.getAttributeNS(AL_NS,'name'),oldstate, state)
