# Copyright (c) 2000-2001 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""RecipeElement module"""

__revision__ = "$Id: RecipeElement.py,v 1.33 2001/10/08 13:50:06 nico Exp $"

from string import split

from narval.lib import *
from narval.bibal import *

from xml.dom.ext import SplitQName

from ALDomAbstraction import ALElement
from StepElement import StepElement
from TransitionElement import TransitionElement
from Prototype import Prototype

class RecipeElement(ALElement) :
    """RecipeElement XXX"""

    # Element methods overriden ################################################

    def __repr__(self):
        return "<RecipeElement Node at %s: with %d attributes and %d children>" \
               % ( id(self), len(self.attributes), len(self.childNodes) )

    # Init methods #############################################################

    def unlink(self) :
        self.name = None
        self.check = None
        self.chktext = ''
        self.identifiers = None
        self.init = false
    
    def link(self) :
        if self.init : return
        start = self.getAttributeNS(AL_NS,'start_step')
        end = self.getAttributeNS(AL_NS,'end_step')
        self.name = self.getAttributeNS(AL_NS,'name')
        self.identifiers = {}
        for node in self.childNodes :
            id = node.getAttributeNS(AL_NS,'id')
            self.identifiers[id] = node
            if node.__class__ == StepElement :
                node.link()
                if id == start :
                    self.start_step = node
                elif id == end :
                    self.end_step = node
            elif node.__class__ == TransitionElement :
                node.link()
        self.id_counter = 0
        self.init = true

    # Recipe element methods ###################################################
    
    def make_plan(self) :
        plan = self.ownerDocument.createElementNS(AL_NS,'plan')
        
        for attr in self.attributes :
            nsuri,qname = SplitQName(attr.name)
            plan.setAttributeNS(nsuri,qname,attr.value)
        for child in self.childNodes :
            if child.nodeType == child.ELEMENT_NODE:
                plan.appendChild(child.cloneNode(1))

        plan.link()
        
        if not plan.getAttributeNS(AL_NS,'decay') :
            plan.setAttributeNS(AL_NS,'decay','90')
            # default DECAY set to 90 seconds, because we use too much RAM

        plan.removeAttributeNS(AL_NS,'eid')
        plan.setAttributeNS(AL_NS,'state','ready')

        return plan

    
    def check_syntax(self) :
        # if that recipe was checked, return result
        if self.check in ['ok', 'bad'] :
            return self.check, self.chktext

        # get name
        recipeName = self.getAttributeNS(AL_NS,'name')
                
        # make lists
        ##transition_list = select_many(self, 'transition')
        ##step_list = select_many(self, 'step')
        transition_list = []
        step_list = []
        transition_ids = {}
        step_ids = {}
        for c in self.childNodes:
            name = c.nodeName
            if name == 'transition':
                transition_list.append(c)
                id = c.getAttributeNS(AL_NS,'id')
                if not id:
                    self.check = 'bad'
                    self.chktext = self.chktext + '\n* Missing transition id.'
                elif transition_ids.has_key(id) or step_ids.has_key(id):
                    self.check = 'bad'
                    self.chktext = self.chktext + '\n* ' + \
                                   'Duplicate transition id %s.' % id
                else:
                    transition_ids[id]=c
                
            elif name == 'step':
                step_list.append(c)
                id = c.getAttributeNS(AL_NS,'id')
                if not id:
                    self.check = 'bad'
                    self.chktext = self.chktext + '\n* Missing step id.'
                elif transition_ids.has_key(id) or step_ids.has_key(id):
                    self.check = 'bad'
                    self.chktext = self.chktext + '\n* ' + \
                                   'Duplicate step id %s.' % id
                else:
                    step_ids[id]=c


        # check start step
        ss_id = self.getAttributeNS(AL_NS,'start_step')
        if (not ss_id) or (ss_id and not step_ids.has_key(ss_id)) :
            self.check = 'bad'
            self.chktext = self.chktext + '\n* Missing start step.'

        # check end_step
        es_id = self.getAttributeNS(AL_NS,'end_step')
        if (not es_id) or (es_id and not step_ids.has_key(es_id)) :
            self.check = 'bad'
            self.chktext = self.chktext + '\n* Missing end step.'

        # check that transitions are pointing to existing steps
        for transition in transition_list :
            inout = []
            for c in transition.childNodes:
                if c.nodeName in ('in','out'):
                    inout.append(c.getAttributeNS(AL_NS,'idref'))
            for ref in inout :
                if not step_ids.has_key(ref):
                    self.check = 'bad'
                    self.chktext = self.chktext + '\n* ' + \
                                   'Bad step reference %s in transition %s' % \
                                   (step.getAttributeNS(AL_NS,'idref'),
                                    transition.getAttributeNS(AL_NS,'id'))

        # check that steps are doing something
        for step in step_list :
            type = step.getAttributeNS(AL_NS,'type')
            a_or_t = None
            if type == 'recipe' :
                pass
            elif type == 'action' :
                # check action exists
                n = step.getAttributeNS(AL_NS,'target')
                if n and not self.ownerDocument.get_action(n):
                    self.check = 'bad'
                    self.chktext = self.chktext + '\n* ' + \
                                   'Unknown action %s for step %s' % \
                                   (step.getAttributeNS(AL_NS,'target'),
                                    step.getAttributeNS(AL_NS,'id'))
            elif type == 'transform' :
                # check that transform exists
                n = step.getAttributeNS(AL_NS,'target')
                if n and not self.ownerDocument.get_transform(n):
                    self.check = 'bad'
                    self.chktext = self.chktext + '\n* ' + \
                                   'Bad transform reference %s in step %s' % \
                                   (step.getAttributeNS(AL_NS,'target'),
                                    step.getAttributeNS(AL_NS,'id'))
            else :
                self.check = 'bad'
                self.chktext = self.chktext + '\n* ' + \
                               'Missing action/recipe attribute for step %s' % \
                               (step.getAttributeNS(AL_NS,"id"))

            # check prototypes
            if a_or_t :
                input_prototype = a_or_t.get_inputs()
                output_prototype = a_or_t.get_outputs()
                for node in step.childNodes:
                    name = node.nodeName
                    if name == 'input' :
                        input_prototype.add(node)
                    elif name == 'output' :
                        output_prototype.add(node)
                        
                # input_prototype
                try:
                    for entry in input_prototype.entries :
                        if not entry['match'] :
                            if entry['optional']:
                                # entry is optional, and noone provides a match
                                # we remove it from the prototype of the step
                                input_prototype.remove(entry)
                            else:
                                raise Exception('no match for %s' % entry['id'])
                except Exception, ex :
                    self.check = 'bad'
                    self.chktext = self.chktext + '\n* ' + \
                                   'input prototype error %s for step %s' % \
                                   (str(ex),
                                    step.getAttributeNS(AL_NS,'id'))
                # output_prototype                
                try:
                    for entry in output_prototype.entries :
                        if not entry['match'] :
                            if entry['optional']:
                                output_prototype.remove(entry)
                            else:
                                raise Exception('no match for %s' % entry['id'])

                except Exception, ex :
                    self.check = 'bad'
                    self.chktext = self.chktext + '\n* ' + \
                                   'output prototype error %s for step %s' % \
                                   (str(ex),
                                    step.getAttributeNS(AL_NS,'id'))

        # if bad, add recipe name in front
        if self.check == 'bad' :
            self.chktext = 'Bad recipe ' + recipeName + ':' + self.chktext
            
        # return result
        return self.check, self.chktext


    # Edit interface ###########################################################

    def generate_new_id(self):
        while 1:
            id = str(self.id_counter)
            self.id_counter = self.id_counter+1
            if not self.identifiers.has_key(id):
                return id

            
    def set_start_step(self,step_node):
        self.start_step = step_node
        self.setAttributeNS(AL_NS,'start_step',
                            step_node.getAttributeNS(AL_NS,'id'))


    def set_end_step(self,step_node):
        self.end_step = step_node
        self.setAttributeNS(AL_NS,'end_step',
                            step_node.getAttributeNS(AL_NS,'id'))


    def change_step_id(self,step,new_id):
        out_transitions = []
        in_transitions = []
        for t in step.out_transitions or []:
            t.remove_in_step(step)
            step.remove_out_transition(t)
            out_transitions.append(t)
        for t in step.in_transitions or []:
            t.remove_out_step(step)
            step.remove_in_transition(t)
            in_transitions.append(t)
        del self.identifiers[step.getAttributeNS(AL_NS,'id')]
        step.setAttributeNS(AL_NS,'id',new_id)
        self.identifiers[new_id]=step
        for t in out_transitions :
            step.parentNode.connect(new_id,
                                    t.getAttributeNS(AL_NS,'id'))
        for t in in_transitions :
            step.parentNode.connect(t.getAttributeNS(AL_NS,'id'),
                                    new_id)
        

    def remove_transition(self,transition):
        print 'removing transition',transition.getAttributeNS(AL_NS,'id')
        print 'in links'
        for id in [x.value for x in select_many(transition,'in/@idref') or []]:
            print id
            self.identifiers[id].remove_out_transition(transition)
        print 'out links'
        for id in [x.value for x in select_many(transition,'out/@idref') or []]:
            print id
            self.identifiers[id].remove_in_transition(transition)
        self._remove_element(transition)


    def remove_step(self,step):
        step_id=step.getAttributeNS(AL_NS,'id')
        print 'removing step',step_id
        for transition in step.out_transitions:
            transition.remove_in_step(step)
        for transition in step.in_transitions:
            transition.remove_out_step(step)
        self._remove_element(step)


    def add_transition(self,transition):
        transition_id=self._add_element(transition)
        for id in [x.value for x in select_many(transition,'in/@idref') or []]:
            self.identifiers[id].out_transitions.append(transition)
        for id in [x.value for x in select_many(transition,'out/@idref') or []]:
            self.identifiers[id].in_transitions.append(transition)
        return transition_id

    
    def add_step(self,step):
        return self._add_element(step)


    def connect(self,from_id,to_id):
        from_node = self.identifiers[from_id]
        to_node = self.identifiers[to_id]
        if from_node.__class__ == StepElement and \
           to_node.__class__ == TransitionElement:
            step,transition = from_node,to_node
            # check that the transition is not already connected
            # to the element
            if transition.in_steps_ids.has_key(from_id):
                return FALSE
            else:
                node=transition.ownerDocument.createElementNS(AL_NS,'in')
                node.setAttributeNS(AL_NS,'idref',from_id)
                step.out_transitions.append(transition)
                transition.in_steps.append(step)
                transition.in_steps_ids[from_id]=None
                transition.appendChild(node)
        elif to_node.__class__ == StepElement and \
             from_node.__class__ == TransitionElement:
            transition,step = from_node,to_node
            # check that the transition is not already connected
            # to the element
            if transition.in_steps_ids.has_key(from_id):
                return FALSE
            else:
                node=transition.ownerDocument.createElementNS(AL_NS,'out')
                node.setAttributeNS(AL_NS,'idref',to_id)
                step.in_transitions.append(transition)
                transition.out_steps.append(step)
                transition.out_steps_ids[to_id]=None
                transition.appendChild(node)            
        else:
            print "I can't connect these!"
            return FALSE
        return TRUE

    # Private ##################################################################

    def _remove_element(self,node):
        id = node.getAttributeNS(AL_NS,'id')
        del self.identifiers[id]
        self.removeChild(node)
        release_domtree(node)


    def _add_element(self,node):
        id = node.getAttributeNS(AL_NS,'id')
        if not id:
            id = self.generate_new_id()
            #if node.hasAttributeNS(AL_NS,'target'):
            #    id = id + '.'+node.getAttributeNS(AL_NS,'target')
            node.setAttributeNS(AL_NS,'id',id)
        self.appendChild(node)
        self.identifiers[id]=node
        return id
    
