# Copyright (c) 2000-2001 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""StepElement module"""

__revision__ = "$Id: StepElement.py,v 1.64 2001/10/02 15:30:56 syt Exp $"

import copy
import time

from narval.lib import *
from narval.bibal import *
from ALDomAbstraction import ALElement
from ChangeListeners import *
from Tags import is_tagged, tag_element, clear_tags, fromtag
from Match import match_filter

from xml.dom import Node

class StepElement(ALElement, PlanChangeListenerInterface,
                  TransitionChangeListenerInterface) :
    """
    prepare
    ... select_input_candidates()
    ... ---> input_candidates = { input: matches, ... }
    ... select_inputs()
    ... ---> inputs = { input: matches, ... }
    run
    ... make_input_sets()
    ... ---> input_sets = [ {input:matches,...}, ...]
    ... foreach input_set: execute()
    ... ---> output_sets = [ (input_set, output_set), ...]
    end
    ... check_outputs()
    ... ---> outputs
    ... tags, outdates, etc.
    """
 
    # State diagram Definition #################################################

    SDD = {'todo': ['ready', 'failed', 'impossible'],
           'ready': ['running'],
           'running': ['end','failed'],
           'end': ['done','failed','error'],
           'done': ['history'],
           'history': [],
           'impossible': [],
           'failed': [],
           'error': ['history']
           }

    # Element methods overriden ################################################

    def __repr__(self):
        # return "<StepElement Node at %s: %d attributes and %d children>" % \
        #        ( id(self), len(self.attributes), len(self.childNodes) )
        return "<StepElement at %s id='%s' target='%s'>" % \
               ( id(self), self.getAttributeNS(AL_NS,'id'),
                 self.getAttributeNS(AL_NS,'target') )

    # Init methods #############################################################

    def unlink(self):
        self.in_transitions = []
        self.out_transitions = []
        self.args = None
        self.delegate = None
        self.output = []
        self.fired_transition = None
        self.init = false
        self.input_prototype = None
        self.output_prototype = None
        
    def link(self):
        """ link a step to its in and out transitions """
        if self.init : return
        plan = self.parentNode
        id = self.getAttributeNS(AL_NS,'id')
        self.in_transitions = []
        self.out_transitions = []
        for t in plan.childNodes:
            if t.nodeName == 'transition':
                for c in t.childNodes:
                    name = c.nodeName
                    if c.getAttributeNS(AL_NS,'idref') == id:
                        if name == 'in':
                            self.out_transitions.append( t )
                        elif name == 'out':
                            self.in_transitions.append( t )
                        
        self.init = true

    # editing interface ########################################################
    
    def remove_in_transition(self,transition):
        """used by Horn"""
        self.in_transitions.remove(transition)

        
    def remove_out_transition(self,transition):
        """used by Horn"""
        self.out_transitions.remove(transition)


    # Step element methods #####################################################
    
    def prepare(self,transition=None) :
        self.fired_transition = transition
        if not self.delegate :
            self._make_delegate()
        self.delegate.prepare()


    def run(self) :
        type = self.getAttributeNS(AL_NS,'type')
        if type == 'recipe' :
            self.delegate.run()
        else :
            self.ownerDocument.create_thread(target=self.delegate.run)

        
    def end(self) :
        self.delegate.end()

            
    def get_outputs(self) :
        if hasattr(self,'outputs') and self.outputs :
            return copy.copy(self.outputs)
        else :
            return []
    
    # ChangeListener interface #################################################

    def transition_change(self,transition) :
        self._evaluate_transitions()

    def plan_change(self,plan) :
        # if child plan is end or failed, this step is end or failed
        if plan != self.parentNode :
            child_plan_state = plan.getAttributeNS(AL_NS,'state')
            if  child_plan_state == 'end' :
                self._set_state('end')
            elif child_plan_state == 'failed' :
                self._set_state('failed')
    
    # Private ##################################################################
    
    def _set_state(self,state) :
        # -- precondition
        oldstate = self.getAttributeNS(AL_NS,'state')
        if state == oldstate : return
        assert (state in self.SDD[oldstate]), \
               'Bad state diagram transition in step %s from %s to %s' \
               % (self.getAttributeNS(AL_NS,'id'), oldstate,state)

        # change state
        self.setAttributeNS(AL_NS,'state',state)

        # propagate change
        self.ownerDocument.step_change(self)
        if state == 'end' : return # shortcut propagation
        self.parentNode.step_change(self)
        for transition in self.out_transitions :
            transition.step_change(self)

        # -- postcondition
        assert (self.getAttributeNS(AL_NS,'state') == state), \
                'Bad state %s after transition in plan %s from %s to %s' \
                % (self.getAttributeNS(AL_NS,'state'),
                   self.getAttributeNS(AL_NS,'name'),oldstate, state)


    def _make_delegate(self) :
        type = self.getAttributeNS(AL_NS,'type')
        if type == 'recipe' :
            self.delegate = RecipeDelegate(self)
        elif type == 'action' :
            self.delegate = ActionDelegate(self)
        elif type == 'transform' :
            self.delegate = TransformDelegate(self)
        else :
            # should *never* get here !
            print "narval.dom.StepElement.make_delegate: WOEAIDH !"
            self._set_state('failed')


    def _evaluate_transitions(self) :
        for transition in self.out_transitions :
            if transition.getAttributeNS(AL_NS,'state') == 'fired' :
                self._set_state('history')
                return
        if self.in_transitions :
            for transition in self.in_transitions :
                if transition.getAttributeNS(AL_NS,'state') != 'impossible' :
                    return
            self._set_state('impossible')

# Delegate interface ###########################################################

class Delegate :
    """Handles the preparation, execution and termination of a step"""

    def __init__(self,step) :
        self.step = step
        self.target = None
        
    def prepare(self) :
        pass

    def run(self) :
        pass
    
    def end(self) :
        pass

# Recipe Delegate ##############################################################

class RecipeDelegate(Delegate) :
    """recipe delegate"""
    
    def prepare(self) :
        self.step._set_state('ready')

    def run(self) :
        STEP = self.step
        STEP._set_state('running')
        name = STEP.getAttributeNS(AL_NS,'target')
        recipe = STEP.ownerDocument.get_recipe(name)
        if recipe :
            code, msg = recipe.check_syntax()
        else :
            msg = 'Unknown recipe %s' % (name)
        if msg :
            err = STEP.ownerDocument.mk_error('bad recipe',msg)
            fromtag(err,STEP)
            STEP.parentNode.add_elements([err])
            STEP._set_state('failed')
        else :
            STEP.ownerDocument.start_plan( name, STEP.parentNode, STEP, [] )
        
    def end(self) :
        self.step._set_state('done')


# Action/Transform Delegate ####################################################

class DelegateException(Exception) :
    """raised when delegate faces problem"""

    def __init__(self,list) :
        self.list = list

class BadInput(DelegateException) :
    """problem with inputs in prepare()"""
    pass

class BadOutput(DelegateException) :
    """problem with outputs in end()"""
    pass

class ErrorOutput(DelegateException) :
    """found <error/> in outputs"""
    pass


class ATDelegate(Delegate) :
    """Abstract class, defines methods common to Action and Transform delegates"""

    def _get_target(self) :
        raise Exception('This method has to be overridden')
    
    def _execute_target(self,input=None) :
        raise Exception('This method has to be overridden')

    # prepare ##################################################################
    
    def prepare(self) :
        STEP = self.step
        self.target = self._get_target()
        self.input_prototype = self.target.get_inputs()
        self.output_prototype = self.target.get_outputs()

        # prototype
        try:
            for node in STEP.childNodes:
                name =  node.nodeName
                if name == 'input':
                    self.input_prototype.add(node)
                elif name == 'output':
                    self.output_prototype.add(node)
            for entry in self.input_prototype.entries :
                if not entry['match'] and entry['optional'] :
                    self.input_prototype.remove(entry)

            for entry in self.output_prototype.entries :
                if not entry['match'] and entry['optional'] :
                    self.output_prototype.remove(entry)
        except Exception, ex:
            import traceback
            traceback.print_exc()
            STEP._set_state('failed')
            return

        # inputs
        try:
            input_candidates = self._select_input_candidates()
            STEP.input_foreach = self._select_foreach(input_candidates)
            STEP.inputs = self._select_inputs(input_candidates)
            STEP._set_state('ready')
        except BadInput, ex :
            for type, msg in ex.list :
                err = STEP.ownerDocument.mk_error(type,msg)
                fromtag(err,STEP)
                STEP.parentNode.add_elements([err])
            STEP._set_state('failed')


    def _select_input_candidates(self) :
        STEP = self.step
        input_candidates = {}
            
        # fetch arguments
        elements = STEP.ownerDocument.get_elements()
        arguments = []
        for c in STEP.childNodes:
            if c.nodeName == 'arguments':
                arguments = [child for child in c.childNodes if child.nodeType==Node.ELEMENT_NODE]
                break
        for input in self.input_prototype.entries :
            input_candidates[input] = []
            use = input['use']
            for e in elements+arguments :
                if (e.getAttributeNS(AL_NS,'outdated') != 'yes') \
                   and ((not use) or (use and not is_tagged(e,STEP))) \
                   and match_filter(e,input['match']) :
                    input_candidates[input].append(e)
                    
        
        # done
        return input_candidates


    def _select_foreach(self,input_candidates) :
        # look for map, fail if imprecise
        foreach_filter = self.step.getAttributeNS(AL_NS,'foreach')
        foreach = None
        if foreach_filter :
            for input,matches in input_candidates.items() :
                if not matches : continue
                for m in matches :
                    if not match_filter(m,[foreach_filter]) :
                        break
                else:
                    if foreach :
                        raise BadInput( [('bad foreach','imprecise')] )
                    elif input['list'] == 'yes' :
                        raise BadInput( [('bad foreach','applies to a list')] )
                    else :
                        foreach = input
        return foreach


    def _select_inputs(self,input_candidates) :
        STEP = self.step
        selected_inputs = {}

        # prepare useful lists
        arguments = []
        for c in STEP.childNodes:
            if c.nodeName == 'arguments':
                map(arguments.append,
                    [child for child in c.childNodes
                           if child.nodeType == Node.ELEMENT_NODE])

        transition_justify = {}
        if STEP.fired_transition :
            transition_justify = STEP.fired_transition.get_justify()

        prev_steps_elements = []
        if STEP.fired_transition :
            prev_steps_elements = STEP.fired_transition.prev_steps_elements()

        # select inputs
        errors = []
        for input in self.input_prototype.entries :

            args, transitions, prev_steps, plan, parent_plan, memory = [], [], [], [], [], []
            use = input['use']

            # order elements by priority
            for m in input_candidates[input] :
                if m in arguments :
                    args.append(m)
                for l in transition_justify.values() :
                    if m in l :
                        transitions.append(m)
                        break
                if m in prev_steps_elements :
                    prev_steps.append(m)
                if m in STEP.parentNode.elements :
                    plan.append(m)
                if STEP.parentNode.parent_plan :
                    if m in STEP.parentNode.parent_plan.elements :
                        parent_plan.append(m)
            memory = input_candidates[input]
            
            # select
            list = input['list'] or (STEP.input_foreach == input)
            optional = input['optional']

            for l in [args, transitions, prev_steps, plan, parent_plan, memory] :
                if (len(l) == 1) or (len(l) > 0 and list) :
                    selected_inputs[input] = l
                    if use :
                        #map(lambda e,s=self.step: tag_element(e,s), l)
                        [tag_element(e,self.step) for e in l]
                    break
            else :
                if optional :
                    selected_inputs[input] = []
                elif list :
                    print input['match']
                    msg = "Missing mandatory input %s" % st(input['match'])
                    errors.append( ('missing input',msg) )
                elif len(l) :
                    msg = "%d inputs instead of one with prototype= %s. Candidates are %s" % \
                          (len(l), str(input), str(map(st,l)))
                    errors.append( ('undecidable input',msg) )
                else :
                    msg = "No input found with prototype=%s" % (str(input))
                    errors.append( ('missing input',msg) )
        # fail on errors
        if errors :
            raise BadInput( errors )

        # ready to run
        return selected_inputs

    # run ######################################################################

    def run(self) :
        STEP = self.step
        STEP._set_state('running')

        input_sets = []
        output_sets = []

        # make input sets
        if STEP.input_foreach :
            for e in STEP.inputs[STEP.input_foreach] :
                input_set = copy.copy(STEP.inputs)
                input_set[STEP.input_foreach] = [e]
                input_sets.append(input_set)
        else :
            input_sets.append( copy.copy(STEP.inputs) )

        # execute
        for input_set in input_sets :
            output_sets.append( (input_set, self._execute(input_set)) )
        # done
        STEP.output_sets = output_sets
        STEP._set_state('end')

        
    def _execute(self,input_set) :
        try:
            # make args tree
            doc = self.step.parentNode.document
            args = doc.createElementNS('','arguments')
            t = time.localtime(time.time())
            args.setAttributeNS('','yeardays',time.strftime('%j',t))
            args.setAttributeNS('','weekdays',time.strftime('%w',t))
            args.setAttributeNS('','seconds', time.strftime('%S',t))
            args.setAttributeNS('','minutes', time.strftime('%M',t))
            args.setAttributeNS('','hours', time.strftime('%H',t))
            args.setAttributeNS('','monthdays', time.strftime('%d',t))
            args.setAttributeNS('','months', time.strftime('%m',t))
            args.setAttributeNS('','years', time.strftime('%Y',t))
            doc.documentElement.appendChild(args)
            for input,element_list in input_set.items() :
                for element in element_list :
                    node = doc.importNode(element,DEEP_COPY)
                    node.setAttributeNS(AL_NS,'input-id',input['id'])
                    args.appendChild(node)
            args.normalize()

#            print '[ARGS'
#            PrettyPrint(args)
#            print ']'

            # get output and delete args
            output = self._execute_target(args)
            #args.parentNode.removeChild(args)
            release_domtree(args)
        except : # Exception,e: # catch anything
            import traceback
            traceback.print_exc()
            plan = self.step.parentNode
            print 'exception. plan :',plan
##            import pprint
##            pprint.pprint( self.step.parentNode.__dict__)
            error_xml = mk_traceback_error(plan, sys.exc_info())
            error_xml.appendChild(error_xml.ownerDocument.importNode(args,1))
            output = plan.document.createElementNS('','outputs')
            output.appendChild(error_xml)
        return output

    # end ######################################################################

    def end(self) :
        STEP = self.step
        final_state = 'done'
        STEP.outputs = []
        
        # check output
        for input_set, output_tree in STEP.output_sets :
            try:
                if not output_tree :
                    raise BadOutput( [ 'Not a tree' ] )
                STEP.parentNode.document.documentElement.appendChild(output_tree)
                self._check_output_is_correct(output_tree)
                output_ok = self._match_output(output_tree)
                self._accept_output(input_set, output_ok)
                STEP.outputs.append(output_ok)
            except ErrorOutput, ex :
                for error in ex.list :
                    error = STEP.ownerDocument.importNode(error,DEEP_COPY)
                    fromtag(error,STEP)
                    STEP.parentNode.add_elements([error])
                if final_state != 'failed' :
                    final_state = 'error'

            except BadOutput, ex :
                msg = 'Action %s at step %s returned bad output.\n'\
                      'Expected %s, got %s.' % \
                      (STEP.getAttributeNS(AL_NS,'target'),
                       STEP.getAttributeNS(AL_NS,'id'),
                       str(map(str,self.output_prototype.entries)),
                       element_to_string(output_tree))
                for txt in ex.list :
                    msg = msg+'\n---\n'+txt
                error = STEP.ownerDocument.mk_error('bad output',msg)
                STEP.parentNode.add_elements([error])
                final_state = 'failed'

            # clean
            if output_tree :
                if output_tree.parentNode:
                    output_tree.parentNode.removeChild(output_tree)
                release_domtree(output_tree)

        # end
        STEP._set_state(final_state)


    def _check_output_is_correct(self,output_tree) :
        # remove 'eid'
        [node.removeAttributeNS(AL_NS,'eid') for node in output_tree.childNodes]

        # check that we got an output tree
        if output_tree.tagName != 'outputs' :
            raise BadOutput( ['Bad output tree root'] )

        # errors
        errors = [child for child in output_tree.childNodes if child.nodeName=='error']
        if errors :
            raise ErrorOutput( errors )


    def _match_output(self,output_tree) :
        # prepare output_ok
        output_ok = {}
        for output in self.output_prototype.entries :
            output_ok[output] = []
            
        # try to match output elements to prototype
        match_outputs = {}
        problems = []
        for element in output_tree.childNodes :
            match_outputs[element] = [output for output in self.output_prototype.entries
                                             if match_filter(element,output['match'])]
            if len(match_outputs[element]) == 1 :
                output_ok[match_outputs[element][0]].append(element)
            else :
                problems.append( "Can't match output %s" % (st(element)) )
            
        # check that number of outputs is correct
        for output, elements in output_ok.items() :
            if len(elements) == 0 and not output['optional'] :
                problems.append( "Missing output for %s" % st(output) )
            elif len(elements) > 1 and not output['list'] :
                problems.append( "Too many outputs. Got %s expected %s" \
                                   % (st(elements),st(output)) )
        if problems :
            raise BadOutput( problems )

        return output_ok


    def _accept_output(self,input_set,output_ok) :
        STEP = self.step

        for output, element_list in output_ok.items() :
            # import
            imported_elements = [STEP.ownerDocument.importNode(e,DEEP_COPY) for e in element_list]
            output_ok[output] = imported_elements
            
            # tag
            for element in imported_elements :
                fromtag(element,STEP)

            # outdates
            outdates = output['outdates']
            if outdates != '' :
                for input, elements in input_set.items() :
                    if input['id'] == outdates :
                        for e in elements :
                            e.setAttributeNS(AL_NS,'outdated','yes')

            # add to plan
            STEP.parentNode.add_elements(imported_elements)


# Action Delegate ##############################################################

class ActionDelegate(ATDelegate) :
    """Handles the execution of an action on behalf of a step"""

    def _get_target(self) :
        # get action prototype
        STEP = self.step
        target = STEP.ownerDocument.get_action(STEP.getAttributeNS(AL_NS,'target'))
            
        module_name,action_name = target.get_identity()
        import os
        last_modif  = os.stat(config.get_home()+"/modules/%s.py"%module_name)[8]
        last_load = int(target.getAttributeNS(AL_NS,'load_date'))
        if last_modif > last_load:
            target = reloadModule(STEP.ownerDocument,STEP.ownerDocument.add_element,module_name,action_name)
        return target
    
    def _execute_target(self,input) :
        moduleName = self.target.getAttributeNS(AL_NS,'group')
        actionFunc = self.target.getAttributeNS(AL_NS,'func')
        exec('from modules.%s import %s' %(moduleName,actionFunc))
        f = eval(actionFunc)
        return f(input)

# Transform Delegate ###########################################################

class TransformDelegate(ATDelegate) :
    """Handles the execution of a transformation on behalf of a step"""

    def _get_target(self) :
        # get transform prototype
        STEP = self.step
        tn = STEP.getAttributeNS(AL_NS,'target')
        target = STEP.ownerDocument.get_transform(tn)
        group,transform_name = target.get_identity()
        last_modif = os.stat(config.get_home()+'/'+"transforms/%s/%s.xslt"%(group,transform_name))[8]
        last_load = int(target.getAttributeNS(AL_NS,'load_date'))
        if last_modif > last_load:
            target = reloadTransform(STEP.ownerDocument,STEP.ownerDocument.add_element,transform_name,group=group)

        return target
    
    def _execute_target(self,inputs) :
        outputs = self.step.parentNode.document.createElementNS(AL_NS,'outputs')
            
        for e in self.target.transform(inputs,outputs.ownerDocument)[:] :
            if e.nodeType == e.ELEMENT_NODE :
                outputs.appendChild(e)
        return outputs
    
