# Copyright (c) 2000-2001 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""TransitionElement"""

__revision__ = "$Id: TransitionElement.py,v 1.35 2001/08/10 12:14:38 alf Exp $"


from xml.dom.ext import Print

from ALDomAbstraction import ALElement
from ChangeListeners import *
from Tags import *
from Match import match_filter

from narval.lib import *
from narval.bibal import *

import copy

class TransitionElement(ALElement,StepChangeListenerInterface,
                        ElementChangeListenerInterface) :
    """TransitionElement XXX"""
    
    # State diagram Definition #################################################

    SDD = {'wait-step': ['wait-time', 'wait-condition','fireable','impossible'],
           'wait-time': ['wait-condition', 'fireable', 'impossible'],
           'wait-condition': ['fireable', 'impossible'],
           'fireable': ['fired','fail','impossible'],
           'fired': [],
           'fail': [],
           'impossible': []
           }

    # Element methods overriden ################################################

    def __repr__(self):
        return "<TransitionElement Node at %s: %d attributes and %d children>" \
               % ( id(self), len(self.attributes), len(self.childNodes) )

    # ALElement interface ######################################################

    def unlink(self) :
        self.in_steps = []
        self.in_steps_ids = {}
        self.out_steps = []
        self.out_steps_ids = {}
        self.time_conditions = []
        self.conditions = {}
        self.init_conditions = false
        self.justify_table = {}
        self.on_error = {}

    def link(self) :

        # init links
        plan = self.parentNode
        steplinks = {}
        for s in plan.childNodes:
            if s.nodeName == 'step':
                steplinks[s.getAttributeNS(AL_NS,'id')]=s
        
        for node in self.childNodes :
            name = node.nodeName
            if name in ('in','out') :
                ref = node.getAttributeNS(AL_NS,'idref')
                step = steplinks[ref]
                if name == 'in':
                    self.on_error[step] = node.getAttributeNS(AL_NS,'onError')
                    self.in_steps.append(step)
                    self.in_steps_ids[ref] = None
                else:
                    self.out_steps.append( step)
                    self.out_steps_ids[ref] = None
            elif name == 'condition' :
                conds = []
                self.conditions[node] = (conds,[])
                for c in node.childNodes:
                    if c.nodeName == 'match':
                        conds.append(c.firstChild.data)
            elif name == 'time' :
                tc = (node.getAttributeNS(AL_NS,'seconds') or '*',
                      node.getAttributeNS(AL_NS,'minutes') or '*',
                      node.getAttributeNS(AL_NS,'hours') or '*',
                      node.getAttributeNS(AL_NS,'monthdays') or '*',
                      node.getAttributeNS(AL_NS,'months') or '*',
                      node.getAttributeNS(AL_NS,'weekdays') or '*' )
                self.time_conditions.append(tc)
            elif name == 'context':
                pass

    def get_justify(self) :
        return self.justify_table
    
    # editing interface ########################################################
    
    def remove_in_step(self,step):
        id=step.getAttributeNS(AL_NS,'id')
        self.in_steps.remove(step)
        del self.in_steps_ids[id]
        node=select_one(self,'in[@idref="%s"]'%id)
        self.removeChild(node)
        release_domtree(node)

    
    def remove_out_step(self,step):
        id=step.getAttributeNS(AL_NS,'id')
        self.out_steps.remove(step)
        del self.out_steps_ids[id]
        node=select_one(self,'out[@idref="%s"]'%id)
        self.removeChild(node)
        release_domtree(node)

    # others ###################################################################
        
    def fire(self) :
        # tag elements that correspond to 'use' conditions
        for condition, list in self.justify_table.items() :
            if condition.getAttributeNS(AL_NS,'use') == 'yes' :
                for element in list :
                    tag_element(element,self)

        # add to the plan the elements that justify the transition
        for list in self.justify_table.values() :
            self.parentNode.add_elements(list)

        # set new state
        self._set_state('fired')
        
        # run out_steps
        [step.prepare(self) for step in self.out_steps]
            


    def _apply_transform(self, node, list) :
        """deprecated"""
        result = []
        nodeName = node.getAttributeNS(AL_NS,'name')
        if nodeName :
            # apply-transform is a reference to a transform
            transf = self.ownerDocument.get_transform(nodeName)
            for element in list :
                for e in transf.transform(element,self.ownerDocument) :
                    if e.nodeType == node.ELEMENT_NODE :
                        result.append(e)
                    else :
                        print 'TransisitionElement.fire()',e
        else:
            # apply-transform contains a xsl:transform
            for n in node.childNodes :
                if n.nodeName == 'xsl:transform' :
                    memDoc = self.ownerDocument
                    for element in list :
                        result.append(xsl_transform(element,n,memDoc))
                    break # should have only one transform
        # make sure eid can't get thru
        for r in result :
            r.setAttributeNS(AL_NS,'eid','')
        return result
    
    # ChangeListeners interfaces ###############################################

    def step_change(self,step) :
        self._evaluate_steps()

    def element_change(self, element) :
        if self.init_conditions and \
           self.getAttributeNS(AL_NS,'state') == 'wait-condition' :
            for condition in self.conditions.keys() :
                self._evaluate_condition(condition,[element])
            self._evaluate_conditions()

    # time condition ###########################################################
    
    def time_condition_match(self):
        self._set_state('wait-condition')
        self._evaluate_conditions()

    # Private ##################################################################

    def _set_state(self,state) :
        oldstate = self.getAttributeNS(AL_NS,'state')
        if state == oldstate : return
        if state in self.SDD[oldstate] :
            # change state
            self.setAttributeNS(AL_NS,'state',state)

            # propagate change
            self.ownerDocument.transition_change(self)
            self.parentNode.transition_change(self)
            for step in self.out_steps+self.in_steps :
                step.transition_change(self)
        else :
            raise Exception('Bad state diagram transition in Transition' + \
                            'from %s to %s' % (oldstate,state) )


    def _evaluate_steps(self) :
        """Verify state of in-steps
        """
        if self.getAttributeNS(AL_NS,'state') == 'fired' :
            return

        for step in self.in_steps :
            state = step.getAttributeNS(AL_NS,'state')
            on_error = self.on_error[step]

            # if true, transition is impossible
            if (state in ['history','failed','impossible']) \
               or (state == 'done' and on_error == 'yes') \
               or (state == 'error' and on_error != 'yes') :
                self._set_state('impossible')
                return

        for step in self.in_steps :
            state = step.getAttributeNS(AL_NS,'state')
            on_error = self.on_error[step]

            # if true, transition may be possible later: wait
            if (state != 'done' and on_error != 'yes') \
               or (state != 'error' and on_error == 'yes') :
                self._set_state('wait-step')
                return

        # in-steps ok, check time condition
        self._evaluate_time()


    def _evaluate_time(self) :
        """Verify that *at*least*one* of the time condition holds
        """
        if self.time_conditions :
            dates = []
            for time_condition in self.time_conditions :
                sc = shallow_calendar(time_condition)
                dates.append(sc.get_next_date())
            self._set_state('wait-time')
            self.ownerDocument.transition_wait_time(self,min(dates))
        else :
            # no time condition, proceed with other conditions
            self._evaluate_conditions()

        
    def _evaluate_conditions(self) :
        """Check conditions associated with the transition.
        """
        # if not initialized yet, do it now
        if not self.init_conditions :
            elements = self.ownerDocument.get_elements()
            for cond in self.conditions.keys() :
                self._evaluate_condition(cond,elements)
            self.init_conditions = true

        # if at least one condition is not satisfied, return
        for cond, (rules,matches) in self.conditions.items() :
            if not matches :
                self._set_state('wait-condition')
                return

        # every condition is satisfied, select elements
        self._select_elements()

                        
    def _evaluate_condition(self,cond,elements) :
        use = (cond.getAttributeNS(AL_NS,'use') == 'yes')
        rules = self.conditions[cond][0]
        for e in elements :
            if (e.getAttributeNS(AL_NS,'outdated') != 'yes') \
               and ((not use) or (use and not is_tagged(e,self))) \
               and match_filter(e,rules) :
                self.conditions[cond][1].append(e)


    def _select_elements(self) :
        # make useful list
        prev_steps_elements = self.prev_steps_elements()

        # select elements
        for cond, (rules,matches) in self.conditions.items() :
            # order elements by priority
            #steps, plan = [], []
            steps, plan, parent_plan = [], [], []
           
            for m in matches :
                if m in prev_steps_elements :
                    steps.append(m)
                if m in self.parentNode.elements :
                    plan.append(m)

                if self.parentNode.parent_plan :
                    if m in self.parentNode.parent_plan.elements :
                        parent_plan.append(m)

            list = (cond.getAttributeNS(AL_NS,'list') == 'yes')
            context = cond.getAttributeNS(AL_NS,'from')

            if context == 'memory':
                search_list = [matches]
            elif context == 'parent-plan':
                search_list = [parent_plan,matches]
            elif context == 'plan':
                search_list = [plan,parent_plan,matches]
            # '' or 'steps' or ...
            else:
                search_list = [steps,plan,parent_plan,matches]

            for l in search_list :
                # for l in search_list :
                if len(l) > 0 :
                    if list :
                        self.justify_table[cond] = l
                    else :
                        self.justify_table[cond] = [ l[0] ]
                    break
            else :
                print '$$$$$$$$$ Elements disappearing in transitions $$$$$$$$$'

        # ok, transition is fireable
        self._set_state('fireable')

    def prev_steps_elements(self) :
        prev_elements = []
        for s in self.in_steps :
            for step_outputs in s.get_outputs() :
                for output, elements in step_outputs.items() :
                    prev_elements = prev_elements + elements
        return prev_elements
        
