# Copyright (c) 2000-2001 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""SocketManager implements the socket management/proxying service"""

__revision__ = "$Id: SocketManager.py,v 1.17 2001/08/10 12:14:38 alf Exp $"

from asyncore import dispatcher, dispatcher_with_send, loop,socket_map
import socket
import threading
from narval.debug import debug_msg

# this is the only class that should be used outside this module
class SocketManager :
    """Manages sockets. Acts more or less like a port forwarder on a firewall.

    Server Socket
    -------------
    When a server socket is added to the socket manager, it binds to the port,
    starts listening and when someone connects, it accepts the call and get an
    ext_com_sock, then create a socket element and it's matching
    int_srv_sock that's bound to a port and listening.

    When an action connects to int_srv_sock, an int_com_sock is created that
    can read and write to it and everything gets forwarded to ext_srv_sock.

    When a socket element is removed, the matching int_com_sock is removed.
    When a socket-server element is removed the matching srv_sock is removed.


    Client Socket
    -------------
    When a client...

    
    """

    def __init__(self) :
        """__init__() will create the following attributes (they should NEVER
        be accessed directly from the outside) :

        interrupt_sock : an UDP socket used internally to interrupt select()
                          and transmit control messages (type is SOCK_DGRAM)
                    
        interrupt_port : the port on which this socket is bound currently 1242

        select_list : the list to perform the select() call on

        socks : a dictionnary of available sockets indexed by their fileno()s
        
        instructions_stack : a fifo of (function,arguments) tuples
        
        loop : 1 if socket manager should be running, 0 if it should stop

        here are the methods that should be used when dealing with the
        SocketManager class: XXXFIXME
        """
        
        debug_msg('socket_manager','init')

        self.thread=None
        self.loop=0
        self.instructions_stack=[]
        self.socks = []
        
        # interrupt socket
        debug_msg('socket_manager','interrupt socket creation')
        self.interrupt_sock = ctrl_sock(self)
        debug_msg('socket_manager','adding interrupt socket to listen')
        self.interrupt_sock.add_channel()
        
        debug_msg('socket_manager','init over')


    # start / stop /run ########################################################
    
    def start(self) :
        """Use this method to start the socketManager.
        """
        if not self.thread :
            print 'Service','SocketManager','started'
            self.loop = 1
            self.thread = threading.Thread(name='SocketManager',
                                           target=self._run,
                                           args=(lambda x=None: x,))
            self.thread.start()

    def stop(self):
        """use this method to tell the SocketManager to stop running ; but every
        <server-socket> and <socket> elements should have been kill()ed before
        """
        self._control(self._put_loop_to_0)
        self.thread = None # XXXFIXME discutable


    def _run(self,exit_f) :
        """this is the main loop, it is spawn in a separate thread.

        When self._control() sends a packet to interrupt_sock or when data
        arrives on one of the sockets. Copy the data over.
        """

        while self.loop:
            # wait until we have data to read
            loop(timeout=60,use_poll=0)
            

        print "Socket Manager exiting..."

    # create and destroy sockets ###############################################
    
    def new_server_socket(self,port,function,*arguments):
        """use this method to tell the SocketManager to open a socket listening
        on port 'port'. 'function' is a callback the SocketManager will call
        when this is done, which should be used to start the appropriate
        plan and create the <socket> element ; guess what 'arguments' is...
        """
        self._control(self._create_ext_srv_sock,port,function,arguments)

    def new_client_socket(self,hostname,port,sock_element_creator,function,*arguments):
        """use this method to tell the SocketManager to open a connection on
        hostname:port. 'function' is a callback the SocketManager will call
        when this is done, which should be used to start the appropriate
        plan and create the <socket> element ; guess what 'arguments' is...
        """
        self._control(self._create_client_sock,hostname,port,
                      sock_element_creator,
                      function,arguments)
        

    def kill_socket(self,port):
        """use this method to tell the SocketManager to close and remove from
        its tables a socket listening on port 'port' ; this socket may be a
        <server-socket> as well as a <socket>. For client-sockets, port is
        actually (host,port).
        """
        # we use the control method there to avoid killing a socket that has not
        # been created yet
        [self._control(sock.die) for sock in self.socks if sock.port == port]

    # Private ##################################################################
    
    def _control(self,function,*arguments):
        """use this to interrupt the select() call and tell the SocketManager to
        execute 'function' with arguments 'arguments'
        if you want to create a new evenement the SocketManager will react to,
        just create a method containing the evenement's code, then, write a
        wrapper method like stop, new_server_socket and kill, that simply calls
        self.control(the_name_of_your_evenement_method[,arg1[,arg2[,...]]])
        """
        debug_msg('socket_manager','control %s %s'%(str(function),str(arguments)))
        self.instructions_stack.append((function,arguments))
        self.interrupt_sock.sendto('0',
                                   ('localhost',
                                    self.interrupt_sock.interrupt_port))


    def _put_loop_to_0(self):
        """this is used by the stop method to stop the SocketManager's loop
        """
        self.loop=0


    def _create_ext_srv_sock(self,port,function,arguments):
        """this is used by the new_server_socket method to create a new socket
        listening on port port; when it is be done, function 'function' is
        called with arguments 'arguments' (this should be the callback
        function)
        all the subclasses there have a particular role, don't forget the main
        loop calls their do_work method when select returns their fileno
        their die method, when they have one, may also be called by the kill
        method of the SocketManager, or from the die method of another element
        """
        ext_srv = ext_srv_sock(port,self,function,arguments)
        ext_srv.add_channel() # add to listening sockets
        self.socks.append(ext_srv)
        
##    def _create_client_sock(self,hostname,port,sock_element_creator,
##                            function,arguments):
##        """this is used by the new_client_socket method to create a new network
##        connection to hostname:port.
##        listening on port port; when it is be done, function 'function' is
##        called with arguments 'arguments' (this should be the callback
##        function)
##        all the subclasses there have a particular role, don't forget the main
##        loop calls their do_work method when select returns their fileno
##        their die method, when they have one, may also be called by the kill
##        method of the SocketManager, or from the die method of another element
##        """

##        new_sock=cli_sock(hostname,port,self,function,arguments)
##        #self.socks[new_sock.fileno()]=new_sock
##        self.select_list.append(new_sock)
##        sock_element_creator(hostname,port,new_sock.local_port)
##        debug_msg('socket_manager','appended %i, select_list is : %s'%\
##                  (new_sock.fileno(),str(self.select_list)))


# Module Private Classes #######################################################

class ctrl_sock:
    def __init__(self,manager):
        self.sock = socket.socket(socket.AF_INET,socket.SOCK_DGRAM)
        self.sock.setsockopt(socket.SOL_SOCKET,socket.SO_REUSEADDR,1)
        self.interrupt_port = 1242
        self.accepting = 0
        while 1:
            try:
                self.sock.bind(('localhost',self.interrupt_port))
            except:
                self.interrupt_port = self.interrupt_port + 1
                debug_msg('socket_manager','trying next port for interrupt socket')
            else:
                break
        self.sm = manager
        self.fileno = self.sock.fileno
        self.sendto = self.sock.sendto
        self.recv = self.sock.recv

    def writable(self):
        return 0

    def readable(self):
        return 1

    def handle_read_event(self):
        try:
            self.recv(1)
            # take instructions from the instructions stack
            # no need to lock, this is atomical
            function,arguments=self.sm.instructions_stack.pop(0)
            apply(function,arguments)
        except:
            import traceback
            traceback.print_exc()
            
    def add_channel(self, map = None):
        if map is None:
            from asyncore import socket_map
            map = socket_map
        map[self.fileno()]=self

    def del_channel(self, map = None):
        fd = self.fileno()
        if map is None:
            from asyncore import socket_map
            map = socket_map
        map[self.sock.fileno()]=self
        if map.has_key (fd):
            del map [fd]

    def close(self):
        self.del_channel()
        self.sock.close()



class ext_srv_sock(dispatcher):
    """this class corresponds to the <server-socket> element
    """

    def __init__(self,port,socketmanager,function,arguments):
        dispatcher.__init__(self)
        self.cb_function = function
        self.cb_arguments = arguments

        self.sm = socketmanager

        # create sock
        self.port = port
        self.create_socket(socket.AF_INET,socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.bind((socket.gethostname(),port))
        self.listen(5)
        self.int_srv = None
        # debug
        debug_msg('socket_manager', 'created socket, listening on port %i'%port)

    def writable(self):
        return 0

    def __repr__(self):
        return '<ext_srv_sock %d port=%d fileno=%d>'%(id(self),self.port,self.fileno())
    
    def handle_accept(self):
        """Accept connection to get ext_com, then create int_srv and execute
        callback function."""

        debug_msg('socket_manager','accepting connection on port %i'%\
                  (self.getsockname()[1]))

        # accept connection and create socket
        ext_com = com_sock(self.sm, self.accept()[0],master = 1)
        self.int_srv = int_srv_sock(self.sm, ext_com)
        self.int_srv.add_channel()

        # execute callback
        apply(self.cb_function,
              tuple([self.int_srv.port]+list(self.cb_arguments)))


    def handle_close(self):
        self.die()
        
    def die(self):
        """Remove and close socket."""
        debug_msg('socket_manager', '%s dying' % \
                  (repr(self)))
        self.close()
        self.sm.socks.remove(self)
        if self.int_srv:
            self.int_srv.die()

class int_srv_sock(dispatcher):
    """this corresponds to the <socket> element in the plan's memory
    """

    def __init__(self, socketmanager, waiting_sock):
        dispatcher.__init__(self)
        self.waiting_sock = waiting_sock
        waiting_sock.master = self
        self.sm = socketmanager

        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.bind(('localhost',0))
        self.listen(1)
        self.port = self.getsockname()[1]

        debug_msg('socket_manager','listening')

    def __repr__(self):
        return '<int_srv_sock %d port=%d fileno=%d>'%(id(self),self.port,self.fileno())

    def writable(self):
        return 0

    def handle_accept(self):
        """Accept connection and plug result into waiting socket."""
        debug_msg('socket_manager', 'accepting connection on port %i' % \
                  (self.getsockname()[1]))

        # accept connection
        int_com = com_sock(self.sm,self.accept()[0])

        # plug com_sockets into each other
        int_com.plug_dest_dispatcher(self.waiting_sock)
        self.waiting_sock.plug_dest_dispatcher(int_com)
        int_com.add_channel()
        self.waiting_sock.add_channel()


    def handle_close(self):
        self.die()

    def die(self):
        debug_msg('socket_manager', '%s dying' % \
                  (repr(self)))
        if self.waiting_sock:
            self.waiting_sock.die()
        self.close()


class com_sock(dispatcher_with_send):
    """this corresponds to the result of the accept() call performed
    on the sock attribute of the ext_srv_sock or int_srv_class
    """

    def __init__(self, socketmanager, sock, master = None):
        dispatcher_with_send.__init__(self,sock)
        self.sm = socketmanager
        self.dest = None
        self.host,self.port = self.getsockname()
        self.waiting_data = ''
        self.master = master
        
    def __repr__(self):
        return '<com_sock %d host=%s port=%d fileno=%d>'%(id(self),self.host,self.port,self.fileno())

    def handle_read(self):
        """Receive incoming data from sock and copy it to dest."""

        # try reading data
        try:
            incoming_data = self.recv(256)
        except socket.error,why:
            debug_msg('socket_manager','recv : broken pipe')
            self.die()
        else:
            if incoming_data:
                if self.dest:
                    try:
                        self.dest.send(incoming_data)
                    except socket.error,why:
                        self.die()
                else:
                    self.waiting_data = self.waiting_data+incoming_data
            
    def plug_dest_dispatcher(self,dest):
        if self.dest is not None:
            if self.dest.master:
                self.dest.master.die()
            else:
                self.dest.die()
            
        self.dest = dest
        if self.waiting_data:
            self.dest.send(self.waiting_data)
            self.waiting_data = ''

    def unplug_dest_dispatcher(self):
        self.dest = None

    def handle_close(self):
        self.die()

    def die(self):
        debug_msg('socket_manager', '%s dying' % \
                  (repr(self)))
        self.close()
        if self.dest :
            if self.master:
                dest,self.dest = self.dest,None
                dest.die()
            else:
                self.unplug_dest_dispatcher()

##class cli_sock:
##    """this class corresponds to the <client-socket> element
##    """

##    def __init__(self,host,port,socketmanager,function,arguments):
##        sock=socket.socket(socket.AF_INET,socket.SOCK_STREAM)
##        sock.connect((host,port))
##        self.fileno=sock.fileno
##        self.host=host
##        self.port=port
##        self.sock=sock
##        self.dest = None

##        self.sm=socketmanager

##        #socketmanager.ext_socks[(host,port)]=self

##        self.function=function
##        self.arguments=arguments

##        self.append_yourself=self.append_really
##        self.do_work = self.recv
##        self.die=self.die_really
##        debug_msg('socket_manager',
##                  'created socket, connected to %s:%i'%(host,port))

##        new_sock=int_srv_sock(self.sm,self)
##        self.local_port = new_sock.port 
##        #self.sm.socks[new_sock.fileno()]=new_sock
##        self.sm.select_list.append(new_sock.fileno)
##        debug_msg('socket_manager','appended %i, select_list is %s'%\
##                  (new_sock.fileno,str(self.sm.select_list)))
##        self.sm.ext_socks[new_sock.port]=new_sock
##        self.int_srv = new_sock

##    def recv(self):
##        try:
##            incoming_data=self.sock.recv(8192)
##        except socket.error:
##            debug_msg('socket_manager','recv : broken pipe')
##            incoming_data=''
##        if incoming_data:
##            debug_msg('socket_manager','data received on port %i'%\
##                      (self.sock.getsockname()[1]))
##            apply(self.function,tuple([self.local_port]+list(self.arguments)))
##            if self.dest:
##                self.dest.send(incoming_data)
##        else:
##            #FIXME : reconnect ?
##            self.die()

##    def append_really(self,socketmanager):
##        self.append_yourself=self.do_nothing
##        #socketmanager.socks[self.fileno]=self
##        socketmanager.select_list.append(self.fileno)
##        debug_msg('socket_manager',
##                  'appended %i, select_list is : %s'%\
##                  (self.fileno,str(socketmanager.select_list)))

##    def die_really(self):
##        self.die = self.do_nothing
##        self.do_work = self.do_nothing
##        del self.sm.ext_socks[self.fileno]
##        self.sm.select_list.remove(self.fileno)
##        self.sock.close()
##        debug_msg('socket_manager',
##                  'closed %i, select_list is : %s'\
##                  %(self.fileno,str(self.sm.select_list)))

##    def do_nothing(self,foo=None):
##        debug_msg('socket_manager','do_nothing %i'%self.fileno)


# Module Test ##################################################################

if __name__ == '__main__' :
    print "No test function yet..."
